(* Author: Fabian Immler <immler@in.tum.de> *)

theory Auxiliarities
imports Probability
begin

section {* Auxiliarities *}

subsection {* Functions: Injective and Inverse *}

lemma inj_on_vimage_image_eq:
  assumes "inj_on f X" "A \<subseteq> X" shows "f -` f ` A \<inter> X = A"
using assms by (auto simp: vimage_image_eq inj_on_def)

lemma inv_into_inv_into_superset_eq:
  assumes "inj_on f B"
  assumes "bij_betw f A A'" "a \<in> A" "A \<subseteq> B"
  shows "inv_into A' (inv_into B f) a = f a"
proof -
  let ?f' = "inv_into A f" let ?e' = "inv_into B f"
  let ?f'' = "inv_into A' ?f'" let ?e'' = "inv_into A' ?e'"
  have 1: "bij_betw ?f' A' A" using assms by (auto simp add: bij_betw_inv_into)
  obtain a' where 2: "a' \<in> A'" and 3: "?f' a' = a"
    using 1 `a \<in> A` unfolding bij_betw_def by force
  have "f a = a'" using assms 2 3
    by (auto simp add: bij_betw_def)  
  have "inj_on ?e' A'"
  proof (intro inj_onI)
    { fix x assume "x \<in> A'"
      hence "x \<in> f ` A" using assms(2) by (auto simp: bij_betw_def)
      hence "inv_into A f x \<in> A" by (rule inv_into_into)
      also note `A \<subseteq> B`
      finally have "inv_into B f x = ?f' x"
        using f_inv_into_f[OF `x \<in> image f A`]
        by (rule inv_into_f_eq[OF `inj_on f B`])
    }
    moreover
    fix x y assume "x \<in> A'" "y \<in> A'" "inv_into B f x = inv_into B f y"
    ultimately
    have "inv_into A f x = inv_into A f y" by simp
    thus "x = y" by (metis "1" `x \<in> A'` `y \<in> A'` bij_betw_imp_inj_on inj_onD)
  qed
  hence "?e'' a = a'" using assms 2 `f a = a'` by (intro inv_into_f_eq) auto
  thus "?e'' a = f a" using `f a = a'` by simp
qed

lemma f_inv_into_onto:
  fixes f::"'a \<Rightarrow> 'b" and A::"'a set" and B::"'b set"
  assumes "inj_on f A" "B \<subseteq> f ` A"
  shows "f ` inv_into A f ` B = B"
unfolding image_image using assms
proof safe
  fix x assume "x \<in> B"
  thus "x \<in> (\<lambda>x. f (inv_into A f x)) ` B"
    unfolding image_def
    using assms `x \<in> B`
    by (auto simp: Bex_def f_inv_into_f intro!: exI[where x=x])
qed (auto simp: f_inv_into_f)

lemma inj_on_image_subset_iff: "inj_on f (A \<union> B) ==> (f`A <= f`B) = (A<=B)"
by (simp add: inj_on_def, blast)

lemma inv_into_eq:
  assumes "inj_on f A" "inj_on g A"
  assumes "x \<in> g ` A"
  assumes "\<And>i. i \<in> A \<Longrightarrow> f i = g i"
  shows "inv_into A f x = inv_into A g x"
proof -
  from assms obtain y where "g y = x" "y \<in> A" by auto
  show ?thesis
    apply (rule inv_into_f_eq[OF `inj_on f A`])
    apply (rule inv_into_into[OF `x \<in> image g A`])
    apply (subst inv_into_f_eq[OF `inj_on g A`])
    using assms `g y = x` `y \<in> A` by auto
qed

lemma inv_into_eq':
  assumes "inj_on f A" "inj_on f B"
  assumes "x \<in> f ` (A \<inter> B)"
  shows "inv_into A f x = inv_into B f x"
  using assms
  by (metis (full_types) Int_iff f_inv_into_f inv_into_f_f inv_into_into)

subsection {* Topology *}

lemma borel_def_closed: "borel = sigma UNIV (Collect closed)"
  unfolding borel_def
proof (intro sigma_eqI sigma_sets_eqI, safe)
  fix x :: "'a set" assume "open x"
  hence "x = UNIV - (UNIV - x)" by auto
  also have "\<dots> \<in> sigma_sets UNIV (Collect closed)"
    by (rule sigma_sets.Compl)
       (auto intro!: sigma_sets.Basic simp: `open x`)
  finally show "x \<in> sigma_sets UNIV (Collect closed)" by simp
next
  fix x :: "'a set" assume "closed x"
  hence "x = UNIV - (UNIV - x)" by auto
  also have "\<dots> \<in> sigma_sets UNIV (Collect open)"
    by (rule sigma_sets.Compl)
       (auto intro!: sigma_sets.Basic simp: `closed x`)
  finally show "x \<in> sigma_sets UNIV (Collect open)" by simp
qed simp_all

lemma compactE':
  assumes "compact S" "\<forall>n\<ge>m. f n \<in> S"
  obtains l r where "l \<in> S" "subseq r" "((f \<circ> r) ---> l) sequentially"
proof atomize_elim
  have "subseq (op + m)" by (simp add: subseq_def)
  have "\<forall>n. (f o (\<lambda>i. m + i)) n \<in> S" using assms by auto  
  from compactE[OF `compact S` this] guess l r .
  hence "l \<in> S" "subseq ((\<lambda>i. m + i) o r) \<and> (f \<circ> ((\<lambda>i. m + i) o r)) ----> l"
    using subseq_o[OF `subseq (op + m)` `subseq r`] by (auto simp: o_def)
  thus "\<exists>l r. l \<in> S \<and> subseq r \<and> (f \<circ> r) ----> l" by blast
qed

lemma compact_Union [intro]: "finite S \<Longrightarrow> \<forall>T\<in>S. compact T \<Longrightarrow> compact (\<Union>S)"
  by (induct set: finite) auto

lemma closed_UN [intro]: "finite A \<Longrightarrow> \<forall>x\<in>A. compact (B x) \<Longrightarrow> compact (\<Union>x\<in>A. B x)"
  unfolding SUP_def by (rule compact_Union) auto

subsection {* Measures *}

lemma
  UN_finite_countable_eq_Un:
  fixes f :: "'a::countable set \<Rightarrow> _"
  assumes "\<And>s. P s \<Longrightarrow> finite s"
  shows "\<Union>{f s|s. P s} = (\<Union>n::nat. let s = set (from_nat n) in if P s then f s else {})"
proof safe
  fix x X s assume "x \<in> f s" "P s"
  moreover with assms obtain l where "s = set l" using finite_list by auto
  ultimately show "x \<in> (\<Union>n. let s = set (from_nat n) in if P s then f s else {})" using `P s`
    by (auto intro!: exI[where x="to_nat l"])
next
  fix x n assume "x \<in> (let s = set (from_nat n) in if P s then f s else {})"
  thus "x \<in> \<Union>{f s|s. P s}" using assms by (auto simp: Let_def split: split_if_asm)
qed

lemma
  countable_finite_comprehension:
  fixes f :: "'a::countable set \<Rightarrow> _"
  assumes "\<And>s. P s \<Longrightarrow> finite s"
  assumes "\<And>s. P s \<Longrightarrow> f s \<in> sets M"
  shows "\<Union>{f s|s. P s} \<in> sets M"
proof -
  from UN_finite_countable_eq_Un[of P f] assms
  have "\<Union>{f s|s. P s} = (\<Union>n. let s = set (from_nat n) in if P s then f s else {})" by simp
  also have "\<dots> \<in> sets M" using assms by (auto simp: Let_def)
  finally show ?thesis .
qed

lemma (in ring_of_sets) union:
  assumes f: "positive M f" "additive M f" and "A \<in> M" "B \<in> M"
  shows "f (A \<union> B) = f A + f (B - A)"
  using assms by (subst additiveD[OF `additive M f`, symmetric]) auto

lemma (in ring_of_sets) plus:
  assumes f: "positive M f" "additive M f" and "A \<in> M" "B \<in> M"
  shows "f B = f (A \<inter> B) + f (B - A)"
proof -
  have "A \<inter> B \<union> (B - A) = B" by auto
  thus ?thesis using assms
    by (subst additiveD[OF `additive M f`, symmetric]) auto
qed

lemma (in ring_of_sets) union_inter_minus_equality:
  assumes f: "positive M f" "additive M f" and "A \<in> M" "B \<in> M"
  shows "f (A \<union> B) + f (A \<inter> B) + f (B - A) = f A + f B + f (B - A)"
  using union[OF assms] plus[OF assms] by (simp add: ac_simps)

lemma (in ring_of_sets) union_plus_inter_equality:
  assumes f: "positive M f" "additive M f" and "A \<in> M" "B \<in> M"
  shows "f (A \<union> B) + f (A \<inter> B) = f A + f B"
proof cases
  assume "f (B - A) = \<infinity>" hence "f B = \<infinity>" "f (A \<union> B) = \<infinity>"
    using plus[OF assms] union[OF assms] by simp_all
  thus ?thesis by simp
next
  assume "f (B - A) \<noteq> \<infinity>" thus ?thesis using union_inter_minus_equality[OF assms] f assms
    by (subst (asm) ereal_add_cancel_right) (auto dest: positiveD2[where A="B-A"])
qed

lemma emeasure_union_plus_inter_equality:
  assumes "A \<in> sets M" "B \<in> sets M"
  shows "M (A \<union> B) + M (A \<inter> B) = M A + M B"
  by (rule union_plus_inter_equality[OF emeasure_positive emeasure_additive assms])

lemma (in finite_measure) measure_union:
  assumes "A \<in> sets M" "B \<in> sets M"
  shows "measure M (A \<union> B) = measure M A + measure M B - measure M (A \<inter> B)"
  using union_plus_inter_equality[OF emeasure_positive emeasure_additive assms]
  by (simp add: emeasure_eq_measure)

lemma (in ring_of_sets) subtractive:
  assumes f: "positive M f" "additive M f" and "A \<in> M" "B \<in> M" and "A \<subseteq> B"
    and "f A < \<infinity>"
  shows "f (B - A) = f B - f A"
proof -
  note union_inter_minus_equality[OF assms(1-4)]
  moreover have "A \<union> B = B" using assms by auto
  ultimately have "f B = f A + f (B - A)" using assms
    by (subst additiveD[OF `additive M f`, symmetric]) auto
  hence "f B - f A = f A + f (B - A) - f A" using assms by simp
  also have "\<dots> = f (B - A) + f A - f A" using assms by (auto simp: ac_simps)
  also have "\<dots> = f (B - A) + (f A - f A)"
    by (metis ab_semigroup_add_class.add_ac(1) ereal_minus(6) ereal_uminus_uminus)
  also have "f A - f A = 0" using assms by (auto simp: positive_def)
  finally show ?thesis by simp
qed

lemma (in ring_of_sets) subadditive:
  assumes f: "positive M f" "additive M f" and A: "range A \<subseteq> M" and S: "finite S"
  shows "f (\<Union>i\<in>S. A i) \<le> (\<Sum>i\<in>S. f (A i))"
using S
proof (induct S)
  case empty thus ?case using f by (auto simp: positive_def)
next
  case (insert x F)
  hence in_M: "A x \<in> M" "(\<Union> i\<in>F. A i) \<in> M" "(\<Union> i\<in>F. A i) - A x \<in> M" using A by force+
  have subs: "(\<Union> i\<in>F. A i) - A x \<subseteq> (\<Union> i\<in>F. A i)" by auto
  have "(\<Union> i\<in>(insert x F). A i) = A x \<union> ((\<Union> i\<in>F. A i) - A x)" by auto
  hence "f (\<Union> i\<in>(insert x F). A i) = f (A x \<union> ((\<Union> i\<in>F. A i) - A x))"
    by simp
  also have "\<dots> = f (A x) + f ((\<Union> i\<in>F. A i) - A x)"
    using f(2) by (rule additiveD) (insert in_M, auto)
  also have "\<dots> \<le> f (A x) + f (\<Union> i\<in>F. A i)"
    using additive_increasing[OF f] in_M subs by (auto simp: increasing_def intro: add_left_mono)
  also have "\<dots> \<le> f (A x) + (\<Sum>i\<in>F. f (A i))" using insert by (auto intro: add_left_mono)
  finally show "f (\<Union> i\<in>(insert x F). A i) \<le> (\<Sum>i\<in>(insert x F). f (A i))" using insert by simp
qed

lemma  finite_Union:
  fixes A::"'a::countable set"
  assumes "\<And>i. i \<in> A \<Longrightarrow> B i \<in> sigma_sets sp C"
  shows "\<Union>B ` A \<in> sigma_sets sp C"
proof cases
  assume "A = {}" thus ?thesis by (simp add: Empty)
next
  assume "A \<noteq> {}" then obtain a where "a \<in> A" by auto
  have UN: "UNION A B =
    UNION UNIV (\<lambda>i. if from_nat i \<in> A then B (from_nat i) else B a)" using `a \<in> A`
    apply auto
  proof -
    case goal1 thus ?case
    by (auto intro: exI[where x="to_nat xa"])
  next
    case goal2 thus ?case by (auto split: split_if_asm simp add: Bex_def)
  qed
  show ?thesis using assms `a\<in>A` by (auto intro: Union simp: UN)
qed

subsection {* Enumeration of Finite Set *}

definition "enum_finite_max J = (SOME n. \<exists> f. J = f ` {i. i < n} \<and> inj_on f {i. i < n})"

definition enum_finite where
  "enum_finite J =
    (SOME f. J = f ` {i::nat. i < enum_finite_max J} \<and> inj_on f {i. i < enum_finite_max J})"

lemma enum_finite_max:
  assumes "finite J"
  shows "\<exists>f::nat\<Rightarrow>'a. J = f ` {i. i < enum_finite_max J} \<and> inj_on f {i. i < enum_finite_max J}"
    unfolding enum_finite_max_def
    by (rule someI_ex) (rule finite_imp_nat_seg_image_inj_on[OF `finite J`])

lemma enum_finite:
  assumes "finite J"
  shows "J = enum_finite J ` {i::nat. i < enum_finite_max J} \<and>
    inj_on (enum_finite J) {i::nat. i < enum_finite_max J}"
  unfolding enum_finite_def
  by (rule someI_ex[of "\<lambda>f. J = f ` {i::nat. i < enum_finite_max J} \<and>
      inj_on f {i. i < enum_finite_max J}"])
    (rule enum_finite_max[OF `finite J`])

lemma in_set_enum_exist:
  assumes "finite A"
  assumes "y \<in> A"
  shows "\<exists>i. y = enum_finite A i"
  using assms enum_finite by auto

subsection {* Enumeration of Countable Union of Finite Sets *}

locale finite_set_sequence =
  fixes Js::"nat \<Rightarrow> 'a set"
  assumes finite_seq[simp]: "finite (Js n)"
begin

definition set_of_Un where "set_of_Un j = (LEAST n. j \<in> Js n)"

definition index_in_set where "index_in_set J j = (SOME n. j = enum_finite J n)"

definition Un_to_nat where
  "Un_to_nat j = to_nat (set_of_Un j, index_in_set (Js (set_of_Un j)) j)"

lemma inj_on_Un_to_nat:
  shows "inj_on Un_to_nat (\<Union>n::nat. Js n)"
proof (rule inj_onI)
  fix x y
  assume "x \<in> (\<Union>n. Js n)" "y \<in> (\<Union>n. Js n)"
  then obtain ix iy where ix: "x \<in> Js ix" and iy: "y \<in> Js iy" by blast
  assume "Un_to_nat x = Un_to_nat y"
  hence "set_of_Un x = set_of_Un y"
    "index_in_set (Js (set_of_Un y)) y = index_in_set (Js (set_of_Un x)) x"
    by (auto simp: Un_to_nat_def)
  moreover
  have "y \<in> Js (set_of_Un y)" unfolding set_of_Un_def using iy by (rule LeastI)
  have "x \<in> Js (set_of_Un x)" unfolding set_of_Un_def using ix by (rule LeastI)
  have "y = enum_finite (Js (set_of_Un y)) (index_in_set (Js (set_of_Un y)) y)"
    unfolding index_in_set_def
    apply (rule someI_ex) 
    using `y \<in> Js (set_of_Un y)` finite_seq
    apply (auto intro!: in_set_enum_exist)
    done
  moreover have "x = enum_finite (Js (set_of_Un x)) (index_in_set (Js (set_of_Un x)) x)"
    unfolding index_in_set_def
    apply (rule someI_ex) 
    using `x \<in> Js (set_of_Un x)` finite_seq
    apply (auto intro!: in_set_enum_exist)
    done
  ultimately show "x = y" by simp
qed

lemma inj_Un[simp]:
  shows "inj_on (Un_to_nat) (Js n)"
  by (intro subset_inj_on[OF inj_on_Un_to_nat]) (auto simp: assms)

lemma Un_to_nat_injectiveD:
  assumes "Un_to_nat x = Un_to_nat y"
  assumes "x \<in> Js i" "y \<in> Js j"
  shows "x = y"
  using assms
  by (intro inj_onD[OF inj_on_Un_to_nat]) auto

end

subsection {* Sequence of Properties on Subsequences *}

lemma subseq_mono: assumes "subseq r" "m < n" shows "r m < r n"
  using assms by (auto simp: subseq_def)

locale subseqs =
  fixes P::"nat\<Rightarrow>(nat\<Rightarrow>nat)\<Rightarrow>(nat\<Rightarrow>nat)\<Rightarrow>bool"
  assumes ex_subseq: "\<And>n s. subseq s \<Longrightarrow> \<exists>r'. subseq r' \<and> P n s r'"
begin

primrec seqseq where
  "seqseq 0 = id"
| "seqseq (Suc n) = seqseq n o (SOME r'. subseq r' \<and> P n (seqseq n) r')"

lemma seqseq_ex:
  shows "subseq (seqseq n) \<and>
  (\<exists>r'. seqseq (Suc n) = seqseq n o r' \<and> subseq r' \<and> P n (seqseq n) r')"
proof (induct n)
  case 0
  let ?P = "\<lambda>r'. subseq r' \<and> P 0 id r'"
  let ?r = "Eps ?P"
  have "?P ?r" using ex_subseq[of id 0] by (intro someI_ex[of ?P]) (auto simp: subseq_def)
  thus ?case by (auto simp: subseq_def) (simp add: id_def)
next
  case (Suc n)
  then obtain r' where
    Suc': "seqseq (Suc n) = seqseq n \<circ> r'" "subseq (seqseq n)" "subseq r'"
      "P n (seqseq n) r'"
    by blast
  let ?P = "\<lambda>r'a. subseq (r'a ) \<and> P (Suc n) (seqseq n o r') r'a"
  let ?r = "Eps ?P"
  have "?P ?r" using ex_subseq[of "seqseq n o r'" "Suc n"] Suc'
    by (intro someI_ex[of ?P]) (auto intro: subseq_o simp: o_assoc)
  moreover have "seqseq (Suc (Suc n)) = seqseq n \<circ> r' \<circ> ?r"
    by (subst seqseq.simps) (simp only: Suc' o_assoc)
  moreover note subseq_o[OF `subseq (seqseq n)` `subseq r'`]
  ultimately show ?case unfolding Suc' by (auto simp: o_def)
qed

lemma subseq_seqseq:
  shows "subseq (seqseq n)" using seqseq_ex[OF assms] by auto

definition reducer where "reducer n = (SOME r'. subseq r' \<and> P n (seqseq n) r')"

lemma subseq_reducer: "subseq (reducer n)" and reducer_reduces: "P n (seqseq n) (reducer n)"
  unfolding atomize_conj unfolding reducer_def using subseq_seqseq
  by (rule someI_ex[OF ex_subseq])

lemma seqseq_reducer[simp]:
  "seqseq (Suc n) = seqseq n o reducer n"
  by (simp add: reducer_def)

declare seqseq.simps(2)[simp del]

definition diagseq where "diagseq i = seqseq i i"

lemma diagseq_mono: "diagseq n < diagseq (Suc n)"
  unfolding diagseq_def seqseq_reducer o_def
  by (metis subseq_mono[OF subseq_seqseq] less_le_trans lessI seq_suble subseq_reducer)

lemma subseq_diagseq: "subseq diagseq"
  using diagseq_mono by (simp add: subseq_Suc_iff diagseq_def)

primrec fold_reduce where
  "fold_reduce n 0 = id"
| "fold_reduce n (Suc k) = fold_reduce n k o reducer (n + k)"

lemma subseq_fold_reduce: "subseq (fold_reduce n k)"
proof (induct k)
  case (Suc k) from subseq_o[OF this subseq_reducer] show ?case by (simp add: o_def)
qed (simp add: subseq_def)

lemma ex_subseq_reduce_index: "seqseq (n + k) = seqseq n o fold_reduce n k"
  by (induct k) simp_all

lemma seqseq_fold_reduce: "seqseq n = fold_reduce 0 n"
  by (induct n) (simp_all)

lemma diagseq_fold_reduce: "diagseq n = fold_reduce 0 n n"
  using seqseq_fold_reduce by (simp add: diagseq_def)

lemma fold_reduce_add: "fold_reduce 0 (m + n) = fold_reduce 0 m o fold_reduce m n"
  by (induct n) simp_all

lemma diagseq_add: "diagseq (k + n) = (seqseq k o (fold_reduce k n)) (k + n)"
proof -
  have "diagseq (k + n) = fold_reduce 0 (k + n) (k + n)"
    by (simp add: diagseq_fold_reduce)
  also have "\<dots> = (seqseq k o fold_reduce k n) (k + n)"
    unfolding fold_reduce_add seqseq_fold_reduce ..
  finally show ?thesis .
qed

lemma diagseq_sub:
  assumes "m \<le> n" shows "diagseq n = (seqseq m o (fold_reduce m (n - m))) n"
  using diagseq_add[of m "n - m"] assms by simp

lemma subseq_diagonal_rest: "subseq (\<lambda>x. fold_reduce k x (k + x))"
  unfolding subseq_Suc_iff fold_reduce.simps o_def
  by (metis subseq_mono[OF subseq_fold_reduce] less_le_trans lessI add_Suc_right seq_suble
      subseq_reducer)

lemma diagseq_seqseq: "diagseq o (op + k) = (seqseq k o (\<lambda>x. fold_reduce k x (k + x)))"
  by (auto simp: o_def diagseq_add)

lemma eventually_sequentially_diagseq:
  assumes "\<And>n s r. P n s r = (\<forall>i. Q n ((s o r) i))"
  shows "eventually (\<lambda>i. Q n (diagseq i)) sequentially"
  unfolding eventually_sequentially
  apply (intro exI[where x="Suc n"])
  apply safe
  apply (subst diagseq_sub) apply simp
  using reducer_reduces[of n, simplified assms, simplified seqseq_reducer[symmetric]]
  apply simp
  done
  
lemma diagseq_holds:
  assumes seq_property: "\<And>n s r. P n s r = Q n (s o r)"
  assumes subseq_closed: "\<And>n s r. subseq r \<Longrightarrow> Q n s \<Longrightarrow> Q n (s o r)"
  shows "P n diagseq (op + (Suc n))"
  unfolding seq_property diagseq_seqseq
  by (intro subseq_closed subseq_diagonal_rest)
     (auto simp: reducer_reduces seq_property[symmetric])

end

subsection {* Product Sets *}

lemma PiE_def': "Pi\<^isub>E I A = {f. (\<forall>i \<in> I. f i \<in> A i) \<and> f = restrict f I}"
  apply auto
  apply (metis extensional_restrict)
  apply (metis restrict_extensional)
  done

lemma prod_emb_def': "prod_emb I M J X = {a \<in> Pi\<^isub>E I (\<lambda>i. space (M i)). restrict a J \<in> X}"
  by (auto simp: prod_emb_def)

lemma prod_emb_subsetI:
  assumes "F \<subseteq> G"
  shows "prod_emb A M B F \<subseteq> prod_emb A M B G"
  using assms by (auto simp: prod_emb_def)

end
