(* Author: Fabian Immler <immler@in.tum.de> *)

theory Kernels
imports Probability
begin

section{* Kernels *}

text{* Kernels can be used to construct stochastic processes. *}

definition "transfer M K =
  measure_of (space (K undefined)) (sets (K undefined)) (\<lambda>A'. \<integral>\<^isup>+ \<omega>. K \<omega> A' \<partial>M)"

locale kernel =
  fixes M :: "'a measure" and M' :: "'b measure" and K :: "'a \<Rightarrow> 'b measure"
  assumes K_sets[simp]: "\<And>\<omega>. sets (K \<omega>) = sets M'"
  assumes K_measurable: "\<And>A'. A' \<in> sets M' \<Longrightarrow> (\<lambda>\<omega>. emeasure (K \<omega>) A') \<in> borel_measurable M"
begin

abbreviation "\<mu>' \<equiv> transfer M K"

lemma K_space[simp]: "space (K \<omega>) = space M'"
  using sets_eq_imp_space_eq[OF K_sets] .

lemma
  shows space_transfer[simp]: "space \<mu>' = space M'"
  and sets_transfer[simp]: "sets \<mu>' = sets M'"
  by (auto simp: transfer_def)

lemma emeasure_transfer[simp]:
  "A' \<in> sets M' \<Longrightarrow> emeasure \<mu>' A' = (\<integral>\<^isup>+ \<omega>. K \<omega> A' \<partial>M)"
proof (rule emeasure_measure_of[OF transfer_def])
  show "positive (sets \<mu>') (\<lambda>A'. \<integral>\<^isup>+ \<omega>. K \<omega> A' \<partial>M)"
    by (simp add:  transfer_def positive_def positive_integral_positive)
  show "countably_additive (sets \<mu>') (\<lambda>A'. \<integral>\<^isup>+ \<omega>. K \<omega> A' \<partial>M)"
  proof (rule countably_additiveI)
    fix A::"nat \<Rightarrow> 'b set"
    assume A: "range A \<subseteq> sets \<mu>'" "disjoint_family A" "(\<Union>i. A i) \<in> sets \<mu>'"
    then have "(\<Sum>i. (\<integral>\<^isup>+ \<omega>. K \<omega> (A i) \<partial>M)) = (\<integral>\<^isup>+ x. (\<Sum>i. K x (A i)) \<partial>M)"
      by (intro positive_integral_suminf[symmetric]) (auto intro!: K_measurable)
    also have "... = (\<integral>\<^isup>+ \<omega>. K \<omega> (\<Union>i. A i) \<partial>M)"
      using A by (intro positive_integral_cong suminf_emeasure) auto
    finally show "(\<Sum>n. \<integral>\<^isup>+ \<omega>. emeasure (K \<omega>) (A n) \<partial>M) = \<integral>\<^isup>+ \<omega>. emeasure (K \<omega>) (\<Union>i. A i) \<partial>M "
      by simp
  qed
  show "sets (K undefined) \<subseteq> Pow (space (K undefined))" using space_closed by simp
qed simp

end

locale markov_kernel = kernel +
  assumes kernel_space_eq_1: "\<And>\<omega>. emeasure (K \<omega>) (space M') = 1"

sublocale markov_kernel \<subseteq> prob_space "K \<omega>" for \<omega>
  using kernel_space_eq_1 by (intro prob_spaceI) simp

lemma ereal_eq_1[simp]: "ereal x = 1 \<longleftrightarrow> x = 1" by (simp add: one_ereal_def)

